/*
 * Copyright (C) 2021 Thomas Van Acker
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <https://www.gnu.org/licenses/>.
 */

package be.bitscuit.gpslogger;

import androidx.annotation.NonNull;
import androidx.annotation.Nullable;
import androidx.appcompat.app.AppCompatActivity;
import androidx.core.content.ContextCompat;

import android.Manifest;
import android.app.Activity;
import android.content.ComponentName;
import android.content.Context;
import android.content.Intent;
import android.content.ServiceConnection;
import android.content.pm.PackageManager;
import android.net.Uri;
import android.os.Build;
import android.os.Bundle;
import android.os.IBinder;
import android.view.View;
import android.widget.Button;
import android.widget.TextView;
import android.widget.Toast;

import com.codebutchery.androidgpx.data.GPXDocument;
import com.codebutchery.androidgpx.data.GPXRoute;
import com.codebutchery.androidgpx.data.GPXSegment;
import com.codebutchery.androidgpx.data.GPXTrack;
import com.codebutchery.androidgpx.data.GPXTrackPoint;
import com.codebutchery.androidgpx.data.GPXWayPoint;
import com.codebutchery.androidgpx.print.GPXFilePrinter;

import java.io.File;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.net.URI;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

public class MainActivity extends AppCompatActivity implements LocationEventListener, GPXFilePrinter.GPXFilePrinterListener {

    //Static vars
    private static final int STATE_IDLE = 0;
    private static final int STATE_PERMISSION = 1;
    private static final int STATE_LOGGING = 2;
    private static final int STATE_RESULT = 3;

    private static final int REQUEST_CODE_PERMISSION = 100;
    private static final int REQUEST_CODE_SAVE = 101;


    //View vars
    private TextView textInfo;
    private Button buttonExport, buttonStartStop;

    //State vars
    private int state = STATE_IDLE;

    //Location vars
    private ArrayList<LocationEvent> locationEvents;



    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.activity_main);

        //Init views
        textInfo = findViewById(R.id.main_text_info);
        buttonExport = findViewById(R.id.main_button_left);
        buttonStartStop = findViewById(R.id.main_button_right);

        //Add listeners
        buttonExport.setOnClickListener(new View.OnClickListener() {
            @Override
            public void onClick(View view) {
                clickedButtonExport();
            }
        });
        buttonStartStop.setOnClickListener(new View.OnClickListener() {
            @Override
            public void onClick(View view) {
                clickedButtonStartStop();
            }
        });

        //Set text
        buttonExport.setText("Export");
        buttonExport.setEnabled(false);
        buttonStartStop.setText("Start");
        updateInfoText();


        //Init other vars
        locationEvents = new ArrayList<>();
    }


    //Button handlers
    private void clickedButtonExport(){
        //Make sure right state
        if(state != STATE_RESULT) return;

        //Check if enough data
        if(locationEvents.size() < 1){
            Toast.makeText(this, "Not enough locations!", Toast.LENGTH_SHORT).show();
            return;
        }

        //Ask where to save file
        Intent intent = new Intent(Intent.ACTION_CREATE_DOCUMENT);
        intent.addCategory(Intent.CATEGORY_OPENABLE);
        intent.setType("application/gpx");
        intent.putExtra(Intent.EXTRA_TITLE, "GPS_Logger_export.gpx");
        startActivityForResult(intent, REQUEST_CODE_SAVE);

        System.out.println("end buttonExport");
    }

    private void clickedButtonStartStop(){
        //Check state
        switch (state){
            case STATE_IDLE:
            case STATE_PERMISSION:
                //Update state and UI
                state = STATE_PERMISSION;
                updateInfoText();

                //Check permissions (only on Android 6+)
                if(Build.VERSION.SDK_INT >= 23
                        && ContextCompat.checkSelfPermission(this, Manifest.permission.ACCESS_FINE_LOCATION) == PackageManager.PERMISSION_DENIED){
                    //Ask for permission
                    requestPermissions(new String[]{Manifest.permission.ACCESS_FINE_LOCATION}, REQUEST_CODE_PERMISSION);
                    break;
                }

                //Reset/init location storage
                locationEvents = new ArrayList<>();

                //Start LoggingService
                Intent intentStart = new Intent(this, LoggerService.class);
                startService(intentStart);
                Intent intentBind = new Intent(this, LoggerService.class);
                bindService(intentBind, loggerConnection, Context.BIND_AUTO_CREATE);

                //Update UI
                state = STATE_LOGGING;
                updateInfoText();
                buttonStartStop.setText("Stop");

                break;

            case STATE_LOGGING:
                //Get LocationEvents
                locationEvents = loggerService.getLocationEvents();

                //Stop Service
                unbindService(loggerConnection);
                loggerServiceBound = false;
                Intent intent = new Intent(this, LoggerService.class);
                stopService(intent);

                //Update UI
                state = STATE_RESULT;
                updateInfoText();
                buttonStartStop.setText("Reset");
                buttonExport.setEnabled(true);

                break;

            case STATE_RESULT:
                //TODO: add nice dialog for confirmation

                //Reset everything
                locationEvents = new ArrayList<>();

                //Update UI
                state = STATE_IDLE;
                updateInfoText();
                buttonStartStop.setText("Start");
                buttonExport.setEnabled(false);

                break;
        }
    }


    //UI stuff
    private void updateInfoText(){
        System.out.println("Updating text");
        String text = "";

        //State
        text += "Status: ";
        switch (state){
            case STATE_IDLE: text += "IDLE"; break;
            case STATE_PERMISSION: text += "PERMISSION"; break;
            case STATE_LOGGING: text += "LOGGING"; break;
            case STATE_RESULT: text += "RESULT"; break;
        }
        text += "\n";

        //Extra (state-specific)
        switch (state){
            case STATE_IDLE:
                text += "Press START to begin logging!";
                break;
            case STATE_PERMISSION:
                text += "Please grant the permissions.";
                break;
            case STATE_LOGGING:
                text += "Press STOP to end logging.\n\n";

                //Adding location info
                text += "Logged "+locationEvents.size()+" location"+(locationEvents.size()==1?"":"s")+".\n";
                try {
                    text += "Last location (" + locationEvents.get(locationEvents.size() - 1).getLatitude() +
                            ", " + locationEvents.get(locationEvents.size() - 1).getLongitude() + ") received at " +
                            Util.formatMillis(locationEvents.get(locationEvents.size() - 1).getTime(), "HH:mm:ss") + "\n";
                }catch (Exception e){}

                break;
            case STATE_RESULT:
                text += "Press EXPORT to save the data.\n" +
                        "Press RESET to reset the log.\n\n";

                //Adding location info
                text += "Logged "+locationEvents.size()+" location"+(locationEvents.size()==1?"":"s")+".\n";
                try {
                    text += "Last location (" + locationEvents.get(locationEvents.size() - 1).getLatitude() +
                            ", " + locationEvents.get(locationEvents.size() - 1).getLongitude() + ") received at " +
                            Util.formatMillis(locationEvents.get(locationEvents.size() - 1).getTime(), "HH:mm:ss") + "\n\n";
                }catch (Exception e){}
                try {
                    int millis = (int) (locationEvents.get(locationEvents.size()-1).getTime()-locationEvents.get(0).getTime());
                    int h = millis/3600000;
                    int m = (millis%3600000)/60000;
                    int s = ((millis%3600000)%60000)/1000;
                    text += "Total time: "+h+"h "+m+"m "+s+"s\n";
                }catch (Exception e){}
                //TODO: add more metrics

                break;
        }

        //Set text
        final String textCopy = text;
        runOnUiThread(new Runnable() {
            @Override
            public void run() {
                textInfo.setText(textCopy);
            }
        });
    }


    //Permission stuff
    @Override
    public void onRequestPermissionsResult(int requestCode, @NonNull String[] permissions, @NonNull int[] grantResults) {

        System.out.println("onPermissionResult");

        if(requestCode == REQUEST_CODE_PERMISSION){
            if(grantResults.length >= 0 && grantResults[0] == PackageManager.PERMISSION_GRANTED){
                //When granted permission -> start logging
                clickedButtonStartStop();
            }else{
                //When didn't grant permission
                Toast.makeText(this, "Unable to start logging!", Toast.LENGTH_LONG).show();
            }
        }

    }


    //Service stuff
    private LoggerService loggerService;
    private boolean loggerServiceBound = false;
    private ServiceConnection loggerConnection = new ServiceConnection() {
        @Override
        public void onServiceConnected(ComponentName componentName, IBinder iBinder) {
            System.out.println("Connecting to service...");
            //When connected with Service
            LoggerService.LoggerBinder loggerBinder = (LoggerService.LoggerBinder) iBinder;
            loggerService = loggerBinder.getService();
            loggerServiceBound = true;

            //Add locationEvent listener
            loggerService.setLocationEventListener(MainActivity.this);

            //Get LocationEvents up till now
            locationEvents = loggerService.getLocationEvents();
        }

        @Override
        public void onServiceDisconnected(ComponentName componentName) {
            loggerServiceBound = false;
        }
    };


    @Override
    public void onLocationEvent(LocationEvent locationEvent) {
        //Add locationEvent to list
        locationEvents.add(locationEvent);

        //Output to console
        System.out.println("New location!\n" +
                "Lat: " + locationEvent.getLatitude() + "\n" +
                "Long: " + locationEvent.getLongitude() + "\n" +
                "Alt: " + locationEvent.getElevation() + "\n" +
                "Vel: " + locationEvent.getVelocity() + "\n" +
                "Time: " + locationEvent.getTime() + "\n");

        //Update UI
        updateInfoText();
    }


    @Override
    protected void onResume() {
        super.onResume();

        System.out.println("onResume");

        //Bind to service if logging
        if(state == STATE_LOGGING){
            System.out.println("Reconnecting to service");
            Intent intent = new Intent(this, LoggerService.class);
            bindService(intent, loggerConnection, Context.BIND_AUTO_CREATE);
        }
    }

    @Override
    protected void onStop() {
        super.onStop();

        System.out.println("onStop");

        //Unbind from service if logging
        if(state == STATE_LOGGING) {
            unbindService(loggerConnection);
            loggerServiceBound = false;
        }
    }



    //Export GPX stuff
    private void exportAsGPX(PrintStream printStream){
        //Export locationEvents using GPX library
        GPXSegment segment = new GPXSegment();
        System.out.println("Segment created");

        for(int i=0; i<locationEvents.size(); i++){
            LocationEvent event = locationEvents.get(i);

            //Create new point
            GPXTrackPoint point = new GPXTrackPoint((float) event.getLatitude(), (float) event.getLongitude());
            point.setElevation((float) event.getElevation());
            point.setTimeStamp(new Date(event.getTime()));

            //Add point to segment
            segment.addPoint(point);
            System.out.println("Location done " + i);
        }

        //Add segment to track
        GPXTrack track = new GPXTrack();
        track.addSegment(segment);

        System.out.println("track ok");

        //Add track to document
        List<GPXTrack> trackList = new ArrayList<>();
        trackList.add(track);
        GPXDocument document = new GPXDocument(new ArrayList<GPXWayPoint>(), trackList, new ArrayList<GPXRoute>());

        System.out.println("doc ok");

        //Write to file
        document.toGPX(printStream);
        System.out.println("ok");
    }

    @Override
    protected void onActivityResult(int requestCode, int resultCode, @Nullable Intent data) {
        super.onActivityResult(requestCode, resultCode, data);

        System.out.println("onActivityResult");

        //Save file
        if(requestCode == REQUEST_CODE_SAVE && resultCode == Activity.RESULT_OK){
            //Export to file
            try{
                System.out.println("in try");
                Uri uri = data.getData();
                System.out.println("got uri");
                exportAsGPX(new PrintStream(getContentResolver().openOutputStream(uri)));
                System.out.println("export done");
                Toast.makeText(this, "Successfully exported data!", Toast.LENGTH_LONG).show();
            }catch (Exception e){
                System.out.println("Excpetion!\n" + e.getMessage());
                e.printStackTrace();
                Toast.makeText(this, "Error during export:\n"+e.getMessage(), Toast.LENGTH_LONG).show();
            }
        }

    }

    @Override
    public void onGPXPrintStarted() {

    }

    @Override
    public void onGPXPrintCompleted() {
        Toast.makeText(this, "Successfully exported data to file!", Toast.LENGTH_SHORT).show();
    }

    @Override
    public void onGPXPrintError(String message) {
        Toast.makeText(this, "Error during export:\n"+message, Toast.LENGTH_LONG).show();
    }
}